/*
 * main.c
 *
 * Author : Gilbert Brisset, 09/13/2009
 * This program is intended to show how to use OpenGL shaders with some GLX platform :
 *  GLUT platform
 *  gtkgl extension platform
 *  gtkgl area platform
 *  glx native platform
 *
 * It builds a 800 x 400 YUV planar pattern made of 16 x 16 of red, green, blue and black blocks.
 * The three pattern planes as sent as textures in an OpenGl context.
 * YUV to RGB color space conversion is done thanks to a GLSL shader.
 * A service thread is created to count and print how many frames are rendered every second.
 *
 * Known issues :
 *  - it is a rough code demo ; choices are made by code compiling keys rather than command line parameters
 *  - there is no "idle" loop in the GLX native mode
 *  - graphics events (closing window) are not managed so closing the demo leads to errors
 *  - Closing stuff, freeing ressources and error management should be better managed.
 *
 *
 * Compiled with
 *
 * gcc -I/usr/include/gtkgl-2.0 -I/usr/include/gtkglext-1.0 -O3 -Wall -c -fmessage-length=0 `pkg-config --cflags gtk+-2.0` -MMD -MP -MF"main.d" -MT"main.d" -o"main.o" "../main.c"
 * the same for GLee.c and gltools.c
 *
 * Linked with
 *
 * gcc `pkg-config --libs gtk+-2.0` `pkg-config --libs gtkglext-1.0` -o"yuv_rgb_glut_demo"  ./GLee.o ./gltools.o ./main.o   -lpthread -lgtkgl-2.0 -lpng -lrt -lglut
 *
 * The shader (yuv_4XX_to_rgb_shader.fs) should be located in the same directory than the binary.
 *
 * Type ./yuv_rgb_glx_demo to run the demo.
 *
 * Usual disclaimer
 * 	There is no warranty for this program.
 * 	This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <stdio.h>
#include <assert.h>
#include <unistd.h>
#include "gltools.h"   // OpenGL toolkit
#include <pthread.h>

/***********************
 *
 *  GLX framework choice :
 *  you can choose 1 from 4 GLX frame works
 *  - the native GLX api with native X window
 *  - the usual GLUT framework with GLUT window
 *  - the GTKGLAREA framework with GTK window
 *  - the GTGLEXT framework also with GTK window
 *
 *  Also known frameworks not studied here :
 *  - the OpenGL Qt framework
 *  - the LWJGL framework allowing to use OpenGL with Java.
 *
 */
#define USES_GLUT_FRAMEWORK
//#define USES_GTKGLEXT_FRAMEWORK
//#define USES_GTKGLAREA_FRAMEWORK
//#define USES_GLX_FRAMEWORK

#ifdef USES_GTKGLEXT_FRAMEWORK
#	include <gtk-2.0/gtk/gtk.h>
#	include <gtkglext-1.0/gdk/gdkgl.h>
#	include <gtkglext-1.0/gtk/gtkgl.h>
#endif

#ifdef USES_GTKGLAREA_FRAMEWORK
#	include <gtk-2.0/gtk/gtk.h>
#	include <gtkgl/gtkglarea.h>
#endif

#ifdef USES_GLX_FRAMEWORK
#	include <gtk-2.0/gtk/gtk.h>
#endif

/***********************
 *
 *  YUV stuff defines
 *
 *  You can choose 1 from 3 image formats :
 *  - 4:2:0 planar
 *  - 4:2:2 planar
 *  - 4:4:4 planar
 *
 *  You can also define the image size.
 *
 */
#define YUV_420
//#define YUV_422
//#define YUV_444
#define IMAGE_WIDTH      800//640//1920//1280//640//1600
#define IMAGE_HEIGHT     400//480//1080//800//480//1200

#ifdef YUV_420
#define X_INT_FORMAT     2
#define Y_INT_FORMAT     2
#endif

#ifdef YUV_422
#define X_INT_FORMAT     2
#define Y_INT_FORMAT     1
#endif

#ifdef YUV_444
#define X_INT_FORMAT     1
#define Y_INT_FORMAT     1
#endif

#define U_IMAGE_WIDTH    (IMAGE_WIDTH /X_INT_FORMAT)
#define U_IMAGE_HEIGHT   (IMAGE_HEIGHT/Y_INT_FORMAT)
#define V_IMAGE_WIDTH    (IMAGE_WIDTH /X_INT_FORMAT)
#define V_IMAGE_HEIGHT   (IMAGE_HEIGHT/Y_INT_FORMAT)

/***********************
 *
 *  OpenGL textures stuff and defines
 *
 */
#define TEXTURE_TARGET GL_TEXTURE_RECTANGLE_ARB // GL_TEXTURE_2D

#define BIND_Y_IMAGE_TEXTURE     0
#define BIND_U_IMAGE_TEXTURE     1
#define BIND_V_IMAGE_TEXTURE     2
#define OPENGL_TEXTURE_COUNT      3
static  GLuint CameraTextureName [OPENGL_TEXTURE_COUNT] ;

#define SHADER_Y_IMAGE_TEXTURE   0
#define SHADER_U_IMAGE_TEXTURE   1
#define SHADER_V_IMAGE_TEXTURE   2

#define ACTIVE_Y_IMAGE_TEXTURE   (GL_TEXTURE0 + SHADER_Y_IMAGE_TEXTURE)
#define ACTIVE_U_IMAGE_TEXTURE   (GL_TEXTURE0 + SHADER_U_IMAGE_TEXTURE)
#define ACTIVE_V_IMAGE_TEXTURE   (GL_TEXTURE0 + SHADER_V_IMAGE_TEXTURE)

/***********************
 *
 *  OpenGL shader stuff and defines
 *
 */
#define SHADER_FILE_PATH "./"
#define YUV_4XX_TO_WINDOW_SHADER 0
#define SHADERS_COUNT            1
char * ShaderNameTable[SHADERS_COUNT] =
{
   "yuv_4XX_to_rgb_shader.fs",
};

GLuint FSHandleTable[SHADERS_COUNT] ; // shaders files handles
GLuint PSHandleTable[SHADERS_COUNT];  // shaders programs handles
#define MAX_INFO_LOG_SIZE 2048


#ifdef USES_GLUT_FRAMEWORK
int glutWindowId ;
#endif

#ifdef USES_GTKGLAREA_FRAMEWORK
GtkWidget * gl; // le widget opengl
GtkWidget * WindowId ;
GtkWidget* GtkGlArea ;
#endif

#ifdef USES_GTKGLEXT_FRAMEWORK
GtkWidget * GLWidget ; // le widget opengl
GtkWidget * WindowId ;
#endif

#ifdef USES_GLX_FRAMEWORK
Display *     GlxDisplay;
Window        GlxWindow;
GLXContext    GlxContext;
GtkWidget * gl; // le widget opengl
GtkWidget * WindowId ;
#endif

GLboolean npotTexturesAvailable = GL_FALSE;
GLint maxTexSize;                       // maximum allowed size for our 2D textures
GLint maxTexUnits;                      // maximum number of texture image units

static void * YPattern ;
static void * UPattern ;
static void * VPattern ;

GLfloat rgb2cmyMatrix[16] =
{
   -1.0 ,  0.0,  0.0,  0.0,
   0.0 , -1.0,  0.0,  0.0,
   0.0 ,  0.0, -1.0,  0.0,
   1.0 ,  1.0,  1.0,  1.0
};

static int FrameCount, PreviousCount ;

/***********************
 *
 *  Local prototypes
 *
 */
GLchar *LoadShaderText ( const char *fileName ) ;
int  CreateShader  ( int ShaderId ) ;
void EnableShader  ( int ShaderId ) ;
void DisableShader ( void ) ;
void SendIntParam2Shader ( int ShaderId, char *Text, GLint unit ) ;
void SendFloatParam2Shader ( int ShaderId, char *Text, GLfloat Value ) ;
void InitOpenGLTextures ( void ) ;
void CameraDisplay    ( void ) ;
void FillMapWithPattern ( int width, int height, int patternSize ) ;
void ReleaseMap ( void ) ;
int SetOpenGLContext() ;
int SwapBuffers ( void ) ;

/***********************
*  LoadShaderText()
*  Load shader from disk into a null-terminated string
*
*/
GLchar *LoadShaderText ( const char *fileName )
{
   GLchar *shaderText = NULL;
   GLint shaderLength = 0;
   FILE *fp;

   fp = fopen ( fileName, "r" );

   if ( fp != NULL )
   {
      while ( fgetc ( fp ) != EOF )
      {
         shaderLength++;
      }

      rewind ( fp );

      shaderText = ( GLchar * ) malloc ( shaderLength+1 );

      if ( shaderText != NULL )
      {
         fread ( shaderText, 1, shaderLength, fp );
      }

      shaderText[shaderLength] = '\0';

      fclose ( fp );
   }

   return shaderText;
} // end of LoadShaderText()

/**********************************
 * CreateShader()
 *
 * Creates and add a shader in the shaders tables
 * A shader list is a linked list of a shaders  used in one OpenGl Context on several frame buffers
 *
 */
int  CreateShader  ( int ShaderId )
{
   char     fullFileName[255];
   GLchar * fsString;
   const GLchar *fsStringPtr[1];
   GLint success;
   const char *filename ;

   assert ( ShaderId < SHADERS_COUNT ) ;

   filename = ShaderNameTable[ShaderId] ;
// Create shader objects and specify shader text
   sprintf ( fullFileName, "%s%s", SHADER_FILE_PATH, filename );
   printf ( "building %s\n", fullFileName );
   fsString = LoadShaderText ( fullFileName );

   if ( !fsString )
   {
      fprintf ( stderr, "Unable to load \"%s\"\n", fullFileName );
      sleep ( 5 );
      exit ( 0 );
   }

   FSHandleTable[ShaderId] = glCreateShader ( GL_FRAGMENT_SHADER );

   fsStringPtr[0] = fsString;
   glShaderSource ( FSHandleTable[ShaderId], 1, fsStringPtr, NULL );
   free ( fsString );

// Compile shaders and check for any errors
   glCompileShader ( FSHandleTable[ShaderId] );
   glGetShaderiv ( FSHandleTable[ShaderId], GL_COMPILE_STATUS, &success );

   if ( !success )
   {
      GLchar infoLog[MAX_INFO_LOG_SIZE];
      glGetShaderInfoLog ( FSHandleTable[ShaderId], MAX_INFO_LOG_SIZE, NULL, infoLog );
      fprintf ( stderr, "Error in fragment shader %s compilation!\n", filename );
      fprintf ( stderr, "Info log: %s\n", infoLog );
      sleep ( 10 );
      exit ( 0 );
   }

   else
   {
      GLchar infoLog[MAX_INFO_LOG_SIZE];
      glGetShaderInfoLog ( FSHandleTable[ShaderId], MAX_INFO_LOG_SIZE, NULL, infoLog );
      fprintf ( stderr, "Fragment shader %s compile info log: %s\n", filename, infoLog );
   }

// Create program object, attach shader, then link
   PSHandleTable[ShaderId] = glCreateProgram();

   glAttachShader ( PSHandleTable[ShaderId], FSHandleTable[ShaderId] );

   glLinkProgram ( PSHandleTable[ShaderId] );

   glGetProgramiv ( PSHandleTable[ShaderId], GL_LINK_STATUS, &success );

   if ( !success )
   {
      GLchar infoLog[MAX_INFO_LOG_SIZE];
      glGetProgramInfoLog ( PSHandleTable[ShaderId], MAX_INFO_LOG_SIZE, NULL, infoLog );
      fprintf ( stderr, "Error in program %s linkage!\n", filename );
      fprintf ( stderr, "Info log: %s\n", infoLog );
      sleep ( 1 );
      exit ( 0 );
   }

   else
   {
      GLchar infoLog[MAX_INFO_LOG_SIZE];
      glGetProgramInfoLog ( PSHandleTable[ShaderId], MAX_INFO_LOG_SIZE, NULL, infoLog );
      fprintf ( stderr, "Program %s link info log: %s\n", filename, infoLog );
   }

   return EXIT_SUCCESS;

} // end of CreateShader()


/**********************************
 * EnableShader()
 *
 *
 */
void EnableShader  ( int ShaderId )
{
   glUseProgram ( PSHandleTable[ShaderId] );
} // end of EnableShader()

/**********************************
 * DisableShader()
 *
 *
 */
void DisableShader ( void )
{
   glUseProgram ( 0 );
} // end of DisableShader()

/**********************************
 * SendIntParam2Shader()
 *
 *
 */
void SendIntParam2Shader ( int ShaderId, char *Text, GLint unit )
{
   int location;

   location = glGetUniformLocation ( PSHandleTable[ShaderId],Text );
   glUniform1i ( location, unit );  /* Bind Ytex to texture unit 0 */
} // end of SendIntParam2Shader()

/**********************************
 * SendFloatParam2Shader()
 *
 *
 */
void SendFloatParam2Shader ( int ShaderId, char *Text, GLfloat Value )
{
   int location;

   location = glGetUniformLocation ( PSHandleTable[ShaderId], Text );
   glUniform1f ( location,Value );
} // end of SendFloatParam2Shader()

/**********************************
 * ExitShader()
 *
 *
 */
int ExitShader ( int ShaderId )
{
   glDeleteProgram ( PSHandleTable[ShaderId] );
   return 0;
} // end of ExitShader()

/**********************************
 *  OpenGlInit()
 *
 *  initializes all ressources needed by OpenGL
 *
 */
int OpenGlInit ( void )
{
   SetOpenGLContext() ;

   // Make sure required functionality is available: cube maps, auto mip gen, etc.

   if ( !GLEE_VERSION_1_4 )
   {
      fprintf ( stderr, "GLEE 1.4 is not available!\n" );
   }

   if ( GLEE_ARB_texture_non_power_of_two )
   {
      npotTexturesAvailable = GL_TRUE;
   }

   else
   {
      fprintf ( stderr, "GL_ARB_texture_non_power_of_two extension is not available!\n" );
      fprintf ( stderr, "Framebuffer effects will be lower resolution (lower quality).\n\n" );
   }

   glGetIntegerv ( GL_MAX_TEXTURE_SIZE, &maxTexSize );

   printf ( "OpenGL version  : %s\n", glGetString ( GL_VERSION ) ) ;
   printf ( "OpenGL vendor   : %s\n", glGetString ( GL_VENDOR ) ) ;
   printf ( "OpenGL renderer : %s\n", glGetString ( GL_RENDERER ) ) ;
   printf ( "GL_MAX_TEXTURE_IMAGE_UNITS        : %d\n", maxTexUnits );
   printf ( "GL_MAX_TEXTURE_SIZE               : %d\n", maxTexSize );
   GLint texSize;
   glGetIntegerv ( GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB, &texSize );
   printf ( "GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB : %d\n", texSize ) ;

   FillMapWithPattern ( IMAGE_WIDTH, IMAGE_HEIGHT, 16 ) ;

   CreateShader ( YUV_4XX_TO_WINDOW_SHADER ) ;
   InitOpenGLTextures() ;

   return ( EXIT_SUCCESS ) ;
}  // end of OpenGlInit()

/**********************************
 *  OpenGlClose()
 *
 *  releases all ressources needed by OpenGL
 *
 */
int OpenGlClose ( void )
{
   SetOpenGLContext() ;
   ExitShader ( YUV_4XX_TO_WINDOW_SHADER ) ;
   ReleaseMap() ;
   return 0 ;
}  // end of OpenGlClose()

/*********************************
 *  InitOpenGLTextures()
 *
 *  Inits textures for camera image display
 *  using "classical" yuv to rgb shader
 *
 */
void InitOpenGLTextures ( void )
{
   glGenTextures ( OPENGL_TEXTURE_COUNT, CameraTextureName ) ;

// Initiates Tab Texture for camera stream YUV
   glPixelStorei ( GL_UNPACK_ALIGNMENT, 4 ) ;
   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
   glEnable ( TEXTURE_TARGET );

   /* Select texture unit Y as the active unit and bind the Y texture. */
   glActiveTexture ( ACTIVE_Y_IMAGE_TEXTURE );
   glBindTexture ( TEXTURE_TARGET, CameraTextureName[BIND_Y_IMAGE_TEXTURE] );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
   glTexEnvf ( GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL );
   glTexImage2D ( TEXTURE_TARGET, 0, GL_LUMINANCE,
                  IMAGE_WIDTH,
                  IMAGE_HEIGHT,
                  0, GL_LUMINANCE, GL_UNSIGNED_BYTE,
                  0 );

   /* Select texture unit U as the active unit and bind the U texture. */
   glActiveTexture ( ACTIVE_U_IMAGE_TEXTURE );
   glBindTexture ( TEXTURE_TARGET, CameraTextureName[BIND_U_IMAGE_TEXTURE] );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
   glTexEnvf ( GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL );
   glTexImage2D ( TEXTURE_TARGET, 0, GL_LUMINANCE,
                  U_IMAGE_WIDTH,
                  U_IMAGE_HEIGHT,
                  0, GL_LUMINANCE, GL_UNSIGNED_BYTE,
                  0 );
   glTexSubImage2D ( TEXTURE_TARGET, 0, 0, 0,
                     V_IMAGE_WIDTH,
                     V_IMAGE_HEIGHT,
                     GL_LUMINANCE, GL_UNSIGNED_BYTE,
                     VPattern ) ;

   /* Select texture unit V as the active unit and bind the V texture. */
   glActiveTexture ( ACTIVE_V_IMAGE_TEXTURE );
   glBindTexture ( TEXTURE_TARGET, CameraTextureName[BIND_V_IMAGE_TEXTURE] );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
   glTexEnvf ( GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL );
   glTexImage2D ( TEXTURE_TARGET, 0, GL_LUMINANCE,
                  V_IMAGE_WIDTH,
                  V_IMAGE_HEIGHT,
                  0, GL_LUMINANCE, GL_UNSIGNED_BYTE,
                  0 );
} // end of InitOpenGLTextures()

/**********************************
 * OpenGlDisplay()
 *
 * Display an YUV planar pattern using using 3 textures
 * 1 for each YUV plane and a YUV to RGB space color converter
 *
 */
void OpenGlDisplay ( void )
{
   GLfloat tx0, tx1, ty0, ty1 ;  // texture corners
   GLfloat wx0, wx1, wy0, wy1 ;  // window corners
   GLint   viewport[4];	      // Where The Viewport Values Will Be Stored

   SetOpenGLContext() ;
   glViewport ( 0, 0, IMAGE_WIDTH, IMAGE_HEIGHT );
   glGetIntegerv ( GL_VIEWPORT, viewport );	// Retrieves The Viewport Values (X, Y, Width, Height)
   glMatrixMode ( GL_COLOR ) ;
   glLoadIdentity() ;
//glLoadMatrixf(rgb2cmyMatrix) ;
   glMatrixMode ( GL_PROJECTION );
   glLoadIdentity();
//glOrtho (0.0, (GLfloat)viewport[2], 0.0, (GLfloat)viewport[3], 	 0.0, 30.0);
   glOrtho ( 0.0, 1.0, 0.0, 1.0, 0.0, 30.0 );
   glMatrixMode ( GL_MODELVIEW ) ;
   glLoadIdentity() ;
   glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   glEnable ( TEXTURE_TARGET );

   EnableShader ( YUV_4XX_TO_WINDOW_SHADER ) ;

   /* Select texture unit Y as the active unit and bind the Y texture. */
   glActiveTexture ( ACTIVE_Y_IMAGE_TEXTURE );
   glMatrixMode ( GL_TEXTURE ) ;
   glLoadIdentity() ;
   SendIntParam2Shader ( YUV_4XX_TO_WINDOW_SHADER, "Ytex", SHADER_Y_IMAGE_TEXTURE );
   glBindTexture ( TEXTURE_TARGET,CameraTextureName[BIND_Y_IMAGE_TEXTURE] );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
   glTexEnvf ( GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL );
   glTexSubImage2D ( TEXTURE_TARGET, 0, 0, 0,
                     IMAGE_WIDTH,
                     IMAGE_HEIGHT,
                     GL_LUMINANCE, GL_UNSIGNED_BYTE,
                     YPattern ) ;

   glActiveTexture ( ACTIVE_U_IMAGE_TEXTURE );
   glMatrixMode ( GL_TEXTURE ) ;
   glLoadIdentity() ;
   SendIntParam2Shader ( YUV_4XX_TO_WINDOW_SHADER, "Utex", SHADER_U_IMAGE_TEXTURE );
   glBindTexture ( TEXTURE_TARGET,CameraTextureName[BIND_U_IMAGE_TEXTURE] );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
   glTexEnvf ( GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL );
   glTexSubImage2D ( TEXTURE_TARGET, 0, 0, 0,
                     U_IMAGE_WIDTH ,
                     U_IMAGE_HEIGHT,
                     GL_LUMINANCE, GL_UNSIGNED_BYTE,
                     UPattern ) ;

   glActiveTexture ( ACTIVE_V_IMAGE_TEXTURE );
   glMatrixMode ( GL_TEXTURE ) ;
   glLoadIdentity() ;
   SendIntParam2Shader ( YUV_4XX_TO_WINDOW_SHADER, "Vtex", SHADER_V_IMAGE_TEXTURE );
   glBindTexture ( TEXTURE_TARGET,CameraTextureName[BIND_V_IMAGE_TEXTURE] );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
   glTexParameteri ( TEXTURE_TARGET,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
   glTexEnvf ( GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL );
   glTexSubImage2D ( TEXTURE_TARGET, 0, 0, 0,
                     V_IMAGE_WIDTH,
                     V_IMAGE_HEIGHT,
                     GL_LUMINANCE, GL_UNSIGNED_BYTE,
                     VPattern ) ;

   SendFloatParam2Shader ( YUV_4XX_TO_WINDOW_SHADER, "XFormat", ( GLfloat ) X_INT_FORMAT ) ;
   SendFloatParam2Shader ( YUV_4XX_TO_WINDOW_SHADER, "YFormat", ( GLfloat ) Y_INT_FORMAT ) ;

   /* Draw image (again and again). */
   tx0 = 0.0 ;
   tx1 = ( GLfloat ) IMAGE_WIDTH  ;
   ty0 = 0.0 ;
   ty1 = ( GLfloat ) IMAGE_HEIGHT ;
//wx0 = 0.0 ; wx1 = (GLfloat)viewport[2];
//wy1 = 0.0 ; wy0 = (GLfloat)viewport[3] ;
   wx0 = 0.0 ;
   wx1 = +1.0 ;
   wy1 = 0.0 ;
   wy0 = +1.0 ;
   glBegin ( GL_QUADS );

   glTexCoord2f ( tx0, ty0 );
   glVertex3f  ( wx0, wy0, -2.0 );

   glTexCoord2f ( tx0, ty1 );
   glVertex3f  ( wx0, wy1, -2.0 );

   glTexCoord2f ( tx1, ty1 );
   glVertex3f  ( wx1, wy1, -2.0 );

   glTexCoord2f ( tx1, ty0 );
   glVertex3f  ( wx1, wy0, -2.0 );

   glEnd();

   DisableShader ();
   glDisable ( TEXTURE_TARGET );

   glColor3f ( 1.0f, 0.0f, 0.0f ) ;  // red
   glLineWidth ( 3.0 );
   glBegin ( GL_LINES );
   glVertex2i ( 320 + 80, 240 ) ;
   glVertex2i ( 320 - 80, 240 ) ;
   glVertex2i ( 320 , 240 + 80 ) ;
   glVertex2i ( 320 , 240 - 80 ) ;
   glEnd() ;

   SwapBuffers() ;
   glFlush();

   FrameCount++ ;

}  // end of OpenGlDisplay()


/*********************************
 * FillMapWithPattern()
 *
 *  Allocates aligned chunks of memory and
 *  create a YUV 4:2:0 or 4:2:2 or 4:4:4 pattern
 *
 */
void FillMapWithPattern ( int width, int height, int patternSize )
{
   int h, w ;
   GLubyte * Ypt, * Upt, * Vpt ;

// Rough coding : allocates a very huge memory to allow 4:4:4 or 4:2:0
   posix_memalign ( &YPattern, sizeof ( void * ), width * height * 2 );

   if ( NULL == YPattern )
   {
      printf ( "%s ln %d pattern malloc fails\n", __FILE__, __LINE__ ) ;
      return ;
   }

   posix_memalign ( &UPattern, sizeof ( void * ), width * height * 2 );

   if ( NULL == UPattern )
   {
      printf ( "%s ln %d pattern malloc fails\n", __FILE__, __LINE__ ) ;
      return ;
   }

   posix_memalign ( &VPattern, sizeof ( void * ), width * height * 2 );

   if ( NULL == VPattern )
   {
      printf ( "%s ln %d pattern malloc fails\n", __FILE__, __LINE__ ) ;
      return ;
   }

// Y Map
   Ypt = YPattern ;

   for ( h = 0 ; h < height ; h ++ )
   {
      Ypt = YPattern +            h * width ;

      for ( w = 0 ; w < width  ; w ++ )
      {
         if ( ( h / patternSize ) % 2 == 0 )
         {
            if ( ( w / patternSize ) % 2 == 0 )
            {
               *Ypt++ =  82 ;  // Red
            }

            else
            {
               *Ypt++ = 144 ;  // Green
            }
         }

         else
         {
            if ( ( w / patternSize ) % 2 == 0 )
            {
               *Ypt++ =  41 ;  // Blue
            }

            else
            {
               *Ypt++ =  16 ;  // Black
            }
         }
      }
   }

//U and V maps

   for ( h = 0 ; h < height / Y_INT_FORMAT  ; h ++ )
   {
      Upt = UPattern + h * width / X_INT_FORMAT ;
      Vpt = VPattern + h * width / X_INT_FORMAT ;

      for ( w = 0 ; w < width / X_INT_FORMAT ; w ++ )
      {
         if ( ( h / ( patternSize / Y_INT_FORMAT ) ) % 2 == 0 )
         {
            if ( ( w / ( patternSize / X_INT_FORMAT ) ) % 2 == 0 )
            {
               *Upt ++  = 240 ;  // Red
               *Vpt ++  =  90 ;  // Red
            }

            else
            {
               *Upt ++  =  34 ;  // Green
               *Vpt ++  =  54 ;  // Green
            }
         }

         else
         {
            if ( ( w / ( patternSize / X_INT_FORMAT ) ) % 2 == 0 )
            {
               *Upt ++  = 110 ;  // Blue
               *Vpt ++  = 240 ;  // Blue
            }

            else
            {
               *Upt ++  = 128 ;  // Black
               *Vpt ++  = 128 ;  // Black
            }
         }
      }
   }
} // end of FillMapWithPattern()

/*********************************
 * ReleaseMap()
 *
 *  free the allocated map
 *
 */
void ReleaseMap ( void )
{
   if ( YPattern != NULL )	free ( YPattern ) ;

   if ( UPattern != NULL )	free ( UPattern ) ;

   if ( VPattern != NULL )	free ( VPattern ) ;
} // end of FillMapWithPattern()


void * CountingThread ( void * attr )
{
   int LastSecondCount ;

   while ( 1 )
   {
      sleep ( 1 );
      LastSecondCount = FrameCount - PreviousCount ;
      PreviousCount = FrameCount ;
      printf ( "FrameCount %d\n", LastSecondCount ) ;
   }

   exit ( 0 );
}



/**********************************
 * SwapBuffers()
 *
 * Encapsulate GLX frameworks swap buffers primitives
 *
 *
 */
int SwapBuffers ( void )
{

#ifdef USES_GLUT_FRAMEWORK
   glutSwapBuffers() ;
#endif

#ifdef USES_GTKGLEXT_FRAMEWORK
   /* recuperation du contexte et de la surface de notre widget */
   GdkGLContext * contexte = gtk_widget_get_gl_context ( GTK_WIDGET ( GLWidget ) );
   GdkGLDrawable * surface = gtk_widget_get_gl_drawable ( GTK_WIDGET ( GLWidget ) );
   /* activation du contexte */

   if ( gdk_gl_drawable_gl_begin ( surface,contexte ) )
   {
      gdk_gl_drawable_swap_buffers ( surface ); // permutation tampons
      gdk_gl_drawable_gl_end ( surface ); // désactivation du contexte
   }

#endif

#ifdef USES_GTKGLAREA_FRAMEWORK
   gtk_gl_area_swap_buffers ( ( GtkGLArea * ) GtkGlArea );

#endif

#ifdef USES_GLX_FRAMEWORK
   glXSwapBuffers ( GlxDisplay, GlxWindow );

#endif
   return ( 0 ) ;

} // end of SwapBuffers()

/**********************************
 * SetOpenGLContext()
 *
 * Encapsulate GLX frameworks set OpenGL context primitives
 *
 *
 */
int SetOpenGLContext()
{
#ifdef USES_GLUT_FRAMEWORK
   glutSetWindow ( glutWindowId ) ;
   return ( 0 ) ;
#endif

#ifdef USES_GTKGLEXT_FRAMEWORK
   GdkGLContext * contexte = gtk_widget_get_gl_context ( GTK_WIDGET ( GLWidget ) );
   GdkGLDrawable * surface = gtk_widget_get_gl_drawable ( GTK_WIDGET ( GLWidget ) );
   /* activation du contexte */

   if ( gdk_gl_drawable_gl_begin ( surface,contexte ) )
      return ( 0 ) ;
   else
      return ( -1 ) ;

#endif

#ifdef USES_GTKGLAREA_FRAMEWORK
   if ( ! gtk_gl_area_make_current ( ( GtkGLArea * ) GtkGlArea ) )
   {
      printf ( "gtk_gl_area_new error\n" ) ;
      return ( -1 ) ;
   }

   else
      return ( 0 ) ;

#endif

#ifdef USES_GLX_FRAMEWORK
   glXMakeCurrent ( GlxDisplay, GlxWindow, GlxContext );

   return ( 0 ) ;

#endif
} // end of SetOpenGLContext()

#if defined( USES_GTKGLAREA_FRAMEWORK) || defined (USES_GTKGLEXT_FRAMEWORK)
static gboolean gtkIdleFunction ( gpointer   data )
{
   OpenGlDisplay() ;
   return ( TRUE ) ;
}  // end of gtkIdleFunction()

#endif

#ifdef USES_GLX_FRAMEWORK
static Bool WaitForNotify ( Display *d, XEvent *e, char *arg )
{
   return ( e->type == MapNotify ) && ( e->xmap.window == ( Window ) arg );
}

#endif

int main ( int argc, char **argv )
{
// creates the frame counting thread
   pthread_t CountingThreadId ;
   pthread_attr_t CountingThreadAttr ;

   pthread_attr_init ( &CountingThreadAttr );
   pthread_create ( &CountingThreadId,
                    &CountingThreadAttr,
                    CountingThread,
                    NULL );


#ifdef USES_GLUT_FRAMEWORK

   glutInit ( &argc, argv ) ;
   glutInitDisplayMode ( GLUT_RGBA | GLUT_DOUBLE ) ;
   glutInitWindowSize ( IMAGE_WIDTH, IMAGE_HEIGHT ) ;
   glutInitWindowPosition ( 0, 10 ) ;
   glutWindowId = glutCreateWindow ( "YUV to RGB shader demo with GLUT" ) ;
   OpenGlInit () ;
   glutIdleFunc ( OpenGlDisplay ) ;
   glutMainLoop() ;

#endif

#ifdef USES_GTKGLEXT_FRAMEWORK
   gtk_init ( &argc, &argv );

   WindowId = gtk_window_new ( GTK_WINDOW_TOPLEVEL );
   gtk_window_set_title ( GTK_WINDOW ( WindowId ), "YUV to RGB shader demo with GTKGLEXT" );
   gtk_container_set_border_width ( GTK_CONTAINER ( WindowId ), 0 );
   gtk_window_set_default_size  ( GTK_WINDOW ( WindowId ),
                                  IMAGE_WIDTH,
                                  IMAGE_HEIGHT ) ;
   gtk_window_move ( GTK_WINDOW ( WindowId ), 0,  20 );

   GtkWidget* MainBox       = gtk_hbox_new ( FALSE, 1 ); /* TRUE if all children are to be given equal space allotments, 1 the number of pixels to place by default between children */
   gtk_container_add ( GTK_CONTAINER ( WindowId ), MainBox );

   if ( FALSE == gdk_gl_init_check  ( &argc, &argv ) )
   {
      printf ( "gdk_gl_init fails\n" );
      exit ( -1 ) ;
   }

   GdkGLConfig * glconfig; // la config qui va aller avec

   static const gint attrList[] =   // les paramètres de la config
   {
//      GDK_GL_USE_GL,
//      GDK_GL_RGBA,
//      GDK_GL_DOUBLEBUFFER,
//      GDK_GL_RED_SIZE, 1,
//      GDK_GL_GREEN_SIZE, 1,
//      GDK_GL_BLUE_SIZE, 1,
      GDK_GL_NONE
   };
   /* création config */
   //glconfig = gdk_gl_config_new ( attrList );
   glconfig = gdk_gl_config_new_by_mode ( GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH | GDK_GL_MODE_DOUBLE );

   if ( NULL == glconfig )
   {
      printf ( "gdk_gl_config_new fails\n" );
      exit ( -1 ) ;
   }

   /* création d'une drawing area qui va recevoir le support Opengl */
   GLWidget = gtk_drawing_area_new();

   /* ajout du support Opengl */
   gtk_widget_set_gl_capability ( GTK_WIDGET ( GLWidget ),glconfig,NULL,TRUE,GDK_GL_RGBA_TYPE );

   gtk_container_add ( GTK_CONTAINER ( MainBox ),GTK_WIDGET ( GLWidget ) ) ;

   gtk_widget_show_all ( WindowId );

   OpenGlInit () ;

   gtk_idle_add ( gtkIdleFunction, NULL );

   gtk_main ();

#endif // #ifdef USES_GTKGLEXT_FRAMEWORK


#ifdef USES_GTKGLAREA_FRAMEWORK
   gtk_init ( &argc, &argv );

   WindowId = gtk_window_new ( GTK_WINDOW_TOPLEVEL );

   gtk_window_set_title ( GTK_WINDOW ( WindowId ), "YUV to RGB shader demo with GTKGLAREA" );

   gtk_container_set_border_width ( GTK_CONTAINER ( WindowId ), 0 );

   gtk_window_set_default_size  ( GTK_WINDOW ( WindowId ),
                                  IMAGE_WIDTH,
                                  IMAGE_HEIGHT ) ;

   gtk_window_move ( GTK_WINDOW ( WindowId ), 0,  20 );

   GtkWidget* MainBox       = gtk_hbox_new ( FALSE, 1 ); /* TRUE if all children are to be given equal space allotments, 1 the number of pixels to place by default between children */

   gtk_container_add ( GTK_CONTAINER ( WindowId ), MainBox );

   /* Attributes list for the GtkGL areas */
   int attrList[]=
   {
      GDK_GL_RGBA, /* RGBA */
      GDK_GL_DOUBLEBUFFER, /* double buffer */
      GDK_GL_RED_SIZE,1, /* 8 bits de rouge (1 octet) */
      GDK_GL_GREEN_SIZE,1, /* 8 bits de vert */
      GDK_GL_BLUE_SIZE,1, /* 8 bits de bleu */
      GDK_GL_NONE /* marqueur de fin */
   };

   int attrList_[] =
   {
      GDK_GL_USE_GL,//GDK_GL_X_RENDERABLE    , True,
      //GDK_GL_DRAWABLE_TYPE   , GDK_GL_WINDOW_BIT,
      //GDK_GL_RENDER_TYPE     ,
      GDK_GL_RGBA, //GDK_GL_RGBA_BIT,
      //GDK_GL_X_VISUAL_TYPE   , GDK_GL_TRUE_COLOR,
      GDK_GL_RED_SIZE        , 8,
      GDK_GL_GREEN_SIZE      , 8,
      GDK_GL_BLUE_SIZE       , 8,
      GDK_GL_ALPHA_SIZE      , 8,
      GDK_GL_DEPTH_SIZE      , 24,
      GDK_GL_STENCIL_SIZE    , 8,
      GDK_GL_DOUBLEBUFFER    , True,
      //GDK_GL_SAMPLE_BUFFERS  , 1,
      //GDK_GL_SAMPLES         , 4,
      None
   };

   GtkGlArea = gtk_gl_area_new ( attrList );

   if ( GtkGlArea == NULL )
   {
      printf ( "gtk_gl_area_new error\n" ) ;
   }

   gtk_box_pack_start ( GTK_BOX ( MainBox ), GtkGlArea, TRUE, TRUE, 0 );

   gtk_widget_show_all ( WindowId );
   OpenGlInit () ;
   gtk_idle_add ( gtkIdleFunction, NULL );
   gtk_main ();
#endif //#ifdef USES_GTKGLAREA_FRAMEWORK


#ifdef USES_GLX_FRAMEWORK
   int           major, minor;
   XVisualInfo * vi;
   Colormap      cmap;
   XSetWindowAttributes swa;
   XEvent        event;

   /* get a connection */
   GlxDisplay = XOpenDisplay ( 0 );

   if ( glXQueryVersion ( GlxDisplay, &major, &minor ) )
   {
      printf ( "GLX Version %d.%d\n", major, minor ) ;
   }

   else
   {
      printf ( "GLX Version fails\n" ) ;
      exit ( -1 ) ;
   }

// Code for GLX version up to 1.2.

   /* Get an appropriate visual */

   if ( minor == 2 )
   {
      int AttributeList[] = { GLX_RGBA,             /* RGBA */
                              GLX_BUFFER_SIZE,  4,
                              GLX_DOUBLEBUFFER,     /* double buffer */
                              GLX_RED_SIZE,     1,  /* 8 bits de rouge (1 octet) */
                              GLX_GREEN_SIZE,   1,  /* 8 bits de vert */
                              GLX_BLUE_SIZE,    1,  /* 8 bits de bleu */
                              GLX_ALPHA_SIZE,   1,
                              GLX_NONE /* marqueur de fin */
                            };
      vi = glXChooseVisual ( GlxDisplay, DefaultScreen ( GlxDisplay ), AttributeList );

      if ( vi == NULL )
      {
         perror ( "glXchooseVisual error" );
         exit ( -1 ) ;
      }
   }

   else if ( minor == 3 )
   {
      int fbcount;
      int AttributeList[] =
      {
         GLX_X_RENDERABLE    , True,
         GLX_DRAWABLE_TYPE   , GLX_WINDOW_BIT,
         GLX_RENDER_TYPE     , GLX_RGBA_BIT,
         GLX_X_VISUAL_TYPE   , GLX_TRUE_COLOR,
         GLX_RED_SIZE        , 8,
         GLX_GREEN_SIZE      , 8,
         GLX_BLUE_SIZE       , 8,
         GLX_ALPHA_SIZE      , 8,
         GLX_DEPTH_SIZE      , 24,
         GLX_STENCIL_SIZE    , 8,
         GLX_DOUBLEBUFFER    , True,
         //GLX_SAMPLE_BUFFERS  , 1,
         //GLX_SAMPLES         , 4,
         None
      };

      GLXFBConfig *fbc = glXChooseFBConfig ( GlxDisplay, DefaultScreen ( GlxDisplay ),
                                             AttributeList, &fbcount );

      if ( !fbc )
      {
         printf ( "Failed to retrieve a framebuffer config\n" );
         exit ( 1 );
      }

      printf ( "Found %d matching FB configs.\n", fbcount );

      // Pick the FB config/visual with the most samples per pixel
      printf ( "Getting XVisualInfos\n" );
      int best_fbc = -1, worst_fbc = -1, best_num_samp = -1, worst_num_samp = 999;

      int i;

      for ( i = 0; i < fbcount; i++ )
      {
         vi = glXGetVisualFromFBConfig ( GlxDisplay, fbc[i] );

         if ( vi )
         {
            int samp_buf, samples;
            glXGetFBConfigAttrib ( GlxDisplay, fbc[i], GLX_SAMPLE_BUFFERS, &samp_buf );
            glXGetFBConfigAttrib ( GlxDisplay, fbc[i], GLX_SAMPLES       , &samples  );

            printf ( "  Matching fbconfig %d, visual ID 0x%2x: SAMPLE_BUFFERS = %d,"
                     " SAMPLES = %d\n",
                     i, ( unsigned int ) vi -> visualid, samp_buf, samples );

            if ( ( best_fbc < 0 ) || ( samp_buf && ( samples > best_num_samp ) ) )
               best_fbc = i, best_num_samp = samples;

            if ( ( worst_fbc < 0 ) || !samp_buf || ( samples < worst_num_samp ) )
               worst_fbc = i, worst_num_samp = samples;
         }

         XFree ( vi );
      }

      // Get a visual
      int fbc_id = best_fbc;

      //int fbc_id = worst_fbc;

      vi = glXGetVisualFromFBConfig ( GlxDisplay, fbc[ fbc_id ]  );

      printf ( "Chosen visual ID = 0x%x\n", ( unsigned int ) vi->visualid );
   }

   /* Create a colormap */
   cmap = XCreateColormap ( GlxDisplay, RootWindow ( GlxDisplay, vi->screen ),vi->visual, AllocNone );

   /* Create a window */
   swa.colormap = cmap;

   swa.border_pixel = 0;

   swa.event_mask = StructureNotifyMask;

   GlxWindow = XCreateWindow ( GlxDisplay, RootWindow ( GlxDisplay, vi->screen ),
                               30, 10,          // position
                               IMAGE_WIDTH, IMAGE_HEIGHT,        //width height
                               2,               // border width
                               vi->depth,       // depth
                               InputOutput,     // class
                               vi->visual,      // visual
                               CWBorderPixel|CWColormap|CWEventMask, // valuemask
                               &swa );          // attributes

   XStoreName ( GlxDisplay, GlxWindow,"YUV to RGB shader demo with GLX" ) ;

   XMapWindow ( GlxDisplay, GlxWindow );

   XIfEvent ( GlxDisplay, &event, WaitForNotify, ( char* ) GlxWindow );

   /* Create a GLX context */
   GlxContext = glXCreateContext ( GlxDisplay, vi, 0, GL_TRUE );  //GL_FALSE);  // FALSE : indirect rendering, TRUE direct rendering

   if ( ! glXIsDirect ( GlxDisplay, GlxContext ) )
   {
      printf ( "Indirect rendering context obtained !\n" ) ;
   }

   else
   {
      printf ( "Direct rendering context obtained !\n" ) ;
   }

   OpenGlInit () ;

   OpenGlDisplay() ;
   sleep ( 10 );
#endif

   sleep ( 10 ) ;


   return ( 0 ) ;
}  // end of ()


